// // ok, go through the standard setup 
sc.stop()
import org.apache.spark.ml.{Pipeline, PipelineModel}
import org.apache.spark.ml.linalg.Vector
import org.apache.spark.sql.Row
import com.datastax.spark.connector._
import org.apache.spark.SparkContext
import org.apache.spark.SparkContext._
import org.apache.spark.SparkConf
import org.apache.spark.sql.DataFrame
import org.apache.spark.ml.classification._
import scala.math.sqrt
import org.apache.spark.ml.linalg.Vectors
import org.apache.spark.ml.linalg.Vectors._
import org.apache.spark.ml.feature.VectorAssembler
import org.apache.spark.ml.tuning.ParamGridBuilder
import org.apache.spark.ml.tuning.CrossValidator
import org.apache.spark.ml.evaluation._

// These next several may be important for converting df dataframes into typed datasets
import org.apache.spark.sql.types.DataTypes
import org.apache.spark.sql.types.StructField
import org.apache.spark.sql.types.StructType
import org.apache.spark.sql.SQLContext

import org.apache.spark.ml.classification.BinaryLogisticRegressionSummary


val conf = new SparkConf(true).set("spark.cassandra.connection.host", "blinchiki.usask.ca").set("spark.cassandra.auth.username", YOUR_USERNAME).set("spark.cassandra.auth.password", YOUR_PASSWORD)
val sc = new SparkContext("local", "test", conf)

val sqlContext = new SQLContext(sc);
import sqlContext.implicits._;
// try this over 30 minute intervals
val timeslotFromTimestampUDF = udf((ts: java.sql.Timestamp) => ts.getTime() / (30*60*1000))
def readAndAnnotateEthicaDataFrameWithTimeslot(ethicaKeyspace:String, ethicaTableName:String) =
    {
    val df = sqlContext.read.format("org.apache.spark.sql.cassandra").options(Map("keyspace" -> ethicaKeyspace, "table" -> ethicaTableName)).load()
    // use the udf defined above to find the timeslot for the responses
    // and create a column with the time of the accelerometer data binned into timeslot
    df.withColumn("timeslot", timeslotFromTimestampUDF($"record_time"))
    }

    
// Survey Responses
val dfTimeslottedSurveyResponses = readAndAnnotateEthicaDataFrameWithTimeslot("ethica_study_73", "survey_responses")
// ok, for convenience, we narrow this to fewer columns, since we don't need all columns
// note that for other uses, may we wish to also include the "resps" column
dfTimeslottedSurveyResponses.printSchema
dfTimeslottedSurveyResponses.cache


// Now some machinery for the (3D) sensor data
val normUDF = udf((x:Float, y:Float, z:Float) => sqrt((x*x + y*y + z*z).toDouble))

val absUDF = udf((x:Double) => scala.math.abs(x))

def aggregate3DDataFrameByUserAndTimeslot(dfTimeslotted:DataFrame, strAggregatedColName:String, strComponentsSuffix:String) = 
    {
    // ok, now group the accelerometer by TIMESLOT and USER_ID
    val xAxisName = s"x_axis_$strComponentsSuffix";
    val yAxisName = s"y_axis_$strComponentsSuffix";
    val zAxisName = s"z_axis_$strComponentsSuffix";
    
    val dfAggregatedBasicMeasures = dfTimeslotted.groupBy("timeslot", "user_id").agg(avg(normUDF($"x_axis", $"y_axis", $"z_axis")), avg($"x_axis") as xAxisName, avg($"y_axis") as yAxisName, avg($"z_axis") as zAxisName).withColumnRenamed("avg(UDF(x_axis, y_axis, z_axis))",strAggregatedColName)
//    dfAggregatedBasicMeasures
    val strColNamePrefix = "abs_" 
    dfAggregatedBasicMeasures.withColumn(strColNamePrefix + xAxisName, absUDF(dfAggregatedBasicMeasures(xAxisName))).withColumn(strColNamePrefix + yAxisName, absUDF(dfAggregatedBasicMeasures(yAxisName))).withColumn(strColNamePrefix + zAxisName, absUDF(dfAggregatedBasicMeasures(zAxisName)))
    }

//def aggregateAndAvgCols(dfWithTimeslotUserCols:DataFrame, colsToAvg:Seq[String])  // fill this in later -- need to derive the right types for handling the arbitrarily-sized agg clause
    
// ACCELEROMETER DATA
val dfTimeslottedAccelKV = aggregate3DDataFrameByUserAndTimeslot(readAndAnnotateEthicaDataFrameWithTimeslot("ethica_study_73", "linear_accel"), "activity", "activity")

// ORIENTATION DATA
// ok, having situated the survey data in clear timeslots above, we go on to aggregate the accelerometer data into timeslots
val dfTimeslottedOrientationKV = aggregate3DDataFrameByUserAndTimeslot(readAndAnnotateEthicaDataFrameWithTimeslot("ethica_study_73", "orientation"), "orientationMag", "orientation")

// GYROSCOPE DATA
// ok, having situated the survey data in clear timeslots above, we go on to aggregate the accelerometer data into timeslots
val dfTimeslottedGyroscopeKV = aggregate3DDataFrameByUserAndTimeslot(readAndAnnotateEthicaDataFrameWithTimeslot("ethica_study_73", "gyroscope"), "gyroscopeMag", "gyroscope")


// GPS DATA
// ok, having situated the survey data in clear timeslots above, we go on to aggregate the accelerometer data into timeslots
val dfTimeslottedGPSKV = readAndAnnotateEthicaDataFrameWithTimeslot("ethica_study_73", "gps").groupBy("timeslot", "user_id").agg(avg($"accu") as "accu", avg($"lat") as "lat", avg($"lon") as "lon", avg($"speed") as "speed")
dfTimeslottedGPSKV.cache

// ok, now the below joins together the survey and accelerometer data
// ok, now join the accelerometer data with the survey response data
val joinColumns = List("user_id", "timeslot");
val dfJoinedSurveySensor = dfTimeslottedSurveyResponses.join(dfTimeslottedGPSKV, usingColumns = joinColumns).join(dfTimeslottedAccelKV, usingColumns = joinColumns).join(dfTimeslottedOrientationKV, usingColumns = joinColumns).join(dfTimeslottedGyroscopeKV, usingColumns = joinColumns)

dfJoinedSurveySensor.cache
dfJoinedSurveySensor.printSchema

// now create a means for dichotomous classification of the data 
val answeredNotAnsweredNumericResponseCategoryUDF = udf((duration:Int) => (duration match { case -2 => 0.0; case -1 => 0.0; case n if (n >= 0) => 1.0; case m => -1.0; }))






// Now apply this to the data

// ok, now go put the data into a format where we can use machine learning algorithms on it

// // this assembler will create a combined features column
val vectorAssembler = new VectorAssembler().setInputCols(Array("activity", "orientationMag", "gyroscopeMag", "speed", "x_axis_orientation", "y_axis_orientation", "z_axis_orientation", "abs_x_axis_orientation", "abs_y_axis_orientation", "abs_z_axis_orientation")).setOutputCol("features")

val dfLabeledWithNoExplicitFeatures = dfJoinedSurveySensor.withColumn("label", answeredNotAnsweredNumericResponseCategoryUDF($"duration"))
// now create a label column that indicates the ground truth situation for this case
val dfLabeled = vectorAssembler.transform(dfLabeledWithNoExplicitFeatures)
dfLabeled.cache



// create some additional support structures
// ok, now we will use a cross-validation strategy, where we train the model on a subset of data

val fractionSubset1 = 0.5
val Array(dfLabeledSubset1, dfLabeledSubset2) = dfLabeled.randomSplit(Array(fractionSubset1, 1-fractionSubset1))

// creator a classifier that we can use many times over, for each of the different classifiers.  
val binaryClassificationEvaluator = new BinaryClassificationEvaluator().setMetricName("areaUnderROC").setRawPredictionCol("rawPrediction").setLabelCol("label")

//
// ok, now create a DECISION TREE classifier
//

val decisionTreeClassifier = new DecisionTreeClassifier().setFeaturesCol("features").setLabelCol("label")
val decisionTreeModelTrainedOnCompleteData = decisionTreeClassifier.fit(dfLabeled) 

// report an overview of the model
decisionTreeModelTrainedOnCompleteData

// now report on model structure
decisionTreeModelTrainedOnCompleteData.toDebugString
// here we apparently don't have a "summary" from which we can extract the predicted data, so we have to transform it directly
val dfPredictedFromDecisionTreeModelTrainedOnCompleteData = decisionTreeModelTrainedOnCompleteData.transform(dfLabeled)

binaryClassificationEvaluator.evaluate(dfPredictedFromDecisionTreeModelTrainedOnCompleteData)

dfPredictedFromDecisionTreeModelTrainedOnCompleteData.show
dfPredictedFromDecisionTreeModelTrainedOnCompleteData.select("label", "prediction").show


val decisionTreeModelTrainedOnSubset1 = decisionTreeClassifier.fit(dfLabeledSubset1) 
val dfPredictedSubset2FromDecisionTreeModelTrainedOnSubset1 = decisionTreeModelTrainedOnSubset1.transform(dfLabeledSubset2)
binaryClassificationEvaluator.evaluate(dfPredictedSubset2FromDecisionTreeModelTrainedOnSubset1)
// ok, now report on the tree
decisionTreeModelTrainedOnSubset1.toDebugString

val decisionTreeModelTrainedOnSubset2 = decisionTreeClassifier.fit(dfLabeledSubset2) 
val dfPredictedSubset1FromDecisionTreeModelTrainedOnSubset2 = decisionTreeModelTrainedOnSubset2.transform(dfLabeledSubset1)
binaryClassificationEvaluator.evaluate(dfPredictedSubset1FromDecisionTreeModelTrainedOnSubset2)
decisionTreeModelTrainedOnSubset2.toDebugString

val decisionTreeClassifierShallow = new DecisionTreeClassifier().setFeaturesCol("features").setLabelCol("label")



// ok, now create a LOGISTIC REGRESSION classifier
//
val logisticRegression = new LogisticRegression().setRegParam(0.2).setMaxIter(100000).setTol(0.001).setStandardization(true)
val logisticRegressionModelTrainedOnCompleteData = logisticRegression.fit(dfLabeled)

logisticRegressionModelTrainedOnCompleteData.coefficients
logisticRegressionModelTrainedOnCompleteData.summary.objectiveHistory

// now evaluate the predictions from the logistic regression model; for this sort of model, we can do so on the prediction sfrom the summary
binaryClassificationEvaluator.evaluate(logisticRegressionModelTrainedOnCompleteData.summary.predictions)

// A more detailed look at the ROC
val evaluationSummary = logisticRegressionModelTrainedOnCompleteData.summary.asInstanceOf[BinaryLogisticRegressionSummary]
evaluationSummary.areaUnderROC
evaluationSummary.roc.show


logisticRegressionModelTrainedOnCompleteData.summary.predictions.select("probability", "prediction", "label").show(50)

import org.apache.spark.ml.feature.StandardScaler
val logisticRegressionPipeline = new Pipeline().setStages(Array(vectorAssembler, new StandardScaler().setInputCol("features"), new LogisticRegression().setRegParam(0.2).setMaxIter(100000).setTol(0.001).setStandardization(true)))
val logisticRegressionModelFromPipeline = logisticRegressionPipeline.fit(dfLabeledWithNoExplicitFeatures)
// having been fit applying the model to a dataset applies the ENTIRE PIPELINE to that dataset!
val dfLogisticRegressionFromPipeline = logisticRegressionModelFromPipeline.transform(dfLabeledWithNoExplicitFeatures)
binaryClassificationEvaluator.evaluate(dfLogisticRegressionFromPipeline)

//
// ok, now create a NAIVE BAYES classifier
//
val vectorAssemblerNonNegative = new VectorAssembler().setInputCols(Array("activity", "orientationMag", "gyroscopeMag", "abs_x_axis_orientation", "abs_y_axis_orientation", "abs_z_axis_orientation")).setOutputCol("features")
val dfLabeledNonNegative = vectorAssemblerNonNegative.transform(dfLabeled)

val naiveBayesModelTrainedOnNonNegativeData = new NaiveBayes().fit(dfLabeledNonNegative)
val dfNaiveBayesModelPredictions = naiveBayesModelTrainedOnNonNegativeData.transform(dfLabeledNonNegative)
dfNaiveBayesModelPredictions.show
dfNaiveBayesModelPredictions.select("label", "prediction").show

binaryClassificationEvaluator.evaluate(dfNaiveBayesModelPredictions)

val standardizingScalerCreator = new StandardScaler().setInputCol("features").setOutputCol("scaledFeatures").setWithMean(true).setWithStd(true)

val standardizingScaler = standardizingScalerCreator.fit(dfLabeledNonNegative)
standardizingScaler.mean
standardizingScaler.std
val dfLabeledNonNegativeStandardized = standardizingScaler.transform(dfLabeledNonNegative)
val naiveBayesModelTrainedOnStandardizedNonNegativeData = new NaiveBayes().fit(dfLabeledNonNegativeStandardized)
val dfNaiveBayesModelPredictionsForStandardizedData = naiveBayesModelTrainedOnStandardizedNonNegativeData.transform(dfLabeledNonNegativeStandardized)
binaryClassificationEvaluator.evaluate(dfNaiveBayesModelPredictionsForStandardizedData)

//val naiveBayesPipeline = new Pipeline().setStages(Array(vectorAssemblerNonNegative,   scalerTransformer, new NaiveBayes()))
val naiveBayesPipeline = new Pipeline().setStages(vectorAssemblerNonNegative,   standardizingScaler, new NaiveBayes())
val naiveBayesModelFromPipeline = naiveBayesPipeline.fit(dfLabeledWithNoExplicitFeatures)

val naiveBayesPipeline2 = new Pipeline().setStages(Array(vectorAssemblerNonNegative,   new StandardScaler().setInputCol("features"), new NaiveBayes()))
val dfNaiveBayesFromPipeline2 = naiveBayesPipeline2.fit(dfLabeledWithNoExplicitFeatures)


// Now for a Neural Network
val arrayCountNodesInLayers = Array(10, 15, 8, 5, 2)    // 10 first layer nodes (one for each feature), 2 intermediate nodes of sizes 8 and 5, and then 2 outputs (one for each class)
val neuralNetworkClassifier = new MultilayerPerceptronClassifier().setMaxIter(10000).setLayers(arrayCountNodesInLayers).setBlockSize(64).setSeed(1)
val neuralNetworkModel = neuralNetworkClassifier.fit(dfLabeled)
val dfNeuralNetworkModelPredictions = neuralNetworkModel.transform(dfLabeled)

val neuralNetworkEvaluator = new MulticlassClassificationEvaluator().setMetricName("accuracy")
neuralNetworkEvaluator.evaluate(dfNeuralNetworkModelPredictions)

// now engage in manual cross-validation with Neural Networks
val neuralNetworkModelTrainedOnSubset1 = neuralNetworkClassifier.fit(dfLabeledSubset1)
val dfNeuralNetworkModelrainedOnSubset1PredictionsOnSubset2 = neuralNetworkModelTrainedOnSubset1.transform(dfLabeledSubset2)
neuralNetworkEvaluator.evaluate(dfNeuralNetworkModelrainedOnSubset1PredictionsOnSubset2)

val neuralNetworkModelTrainedOnSubset2 = neuralNetworkClassifier.fit(dfLabeledSubset2)
val dfNeuralNetworkModelrainedOnSubset2PredictionsOnSubset1 = neuralNetworkModelTrainedOnSubset2.transform(dfLabeledSubset1)
neuralNetworkEvaluator.evaluate(dfNeuralNetworkModelrainedOnSubset2PredictionsOnSubset1)






// now evaluate the predictions from the logit GLM; for this sort of model, we can do so on the prediction sfrom the summary

// ok, now try putting the classifiers into a pipeline


// first for a decision tree classifier
val countFoldsForCrossValidation = 2 // 2 fold cross validation for here
val evaluator = new BinaryClassificationEvaluator()     // uses area under Receiver Operating Characteristic (ROC) curve to judge binary classification
val paramGridForDecisionTreeCrossValidation = new ParamGridBuilder().addGrid(decisionTreeClassifier.maxDepth, Array(3, 5, 7)).build()
val crossValidatorForDecisionTree = new CrossValidator().setEstimator(decisionTreeClassifier).setEvaluator(binaryClassificationEvaluator).setEstimatorParamMaps(paramGridForDecisionTreeCrossValidation).setNumFolds(countFoldsForCrossValidation)

val crossValidationResultsForDecisionTree = crossValidatorForDecisionTree.fit(dfLabeled)
crossValidationResultsForDecisionTree.avgMetrics
val decisionTreeModelFromCrossValidation = crossValidationResultsForDecisionTree.bestModel
decisionTreeModelFromCrossValidation.parent.extractParamMap

val dfPredictedFromCVDTModel = decisionTreeModelFromCrossValidation.transform(dfLabeled)
dfPredictedFromCVDTModel.select("probability", "prediction", "label").show(100)

// now evaluate the cross-validated model
binaryClassificationEvaluator.evaluate(dfPredictedFromCVDTModel)




//now for logistic regression

val paramGridForLogisticRegressionCrossValidation = new ParamGridBuilder().addGrid(logisticRegression.regParam, Array(0.01, 0.05, 0.10, 0.20, 0.40, 0.6, 0.8, 1)).addGrid(logisticRegression.tol, Array(0.001, 0.01, 0.1)).build()
val crossValidatorForLogisticRegression = new CrossValidator().setEstimator(logisticRegression).setEvaluator(evaluator).setEstimatorParamMaps(paramGridForLogisticRegressionCrossValidation).setNumFolds(countFoldsForCrossValidation)

val crossValidationResultsForLogisticRegression = crossValidatorForLogisticRegression.fit(dfLabeled)
crossValidationResultsForLogisticRegression.avgMetrics
val logisticRegressionModelFromCrossValidation = crossValidationResultsForLogisticRegression.bestModel
logisticRegressionModelFromCrossValidation.parent.extractParamMap

val dfPredictedFromCrossValidatedLogisticRegressionModel = logisticRegressionModelFromCrossValidation.transform(dfLabeled)
dfPredictedFromCrossValidatedLogisticRegressionModel.select("probability", "prediction", "label").show(100)

binaryClassificationEvaluator.evaluate(dfPredictedFromCrossValidatedLogisticRegressionModel)



// ok, now prepare for support vector machine (SVM) classifier
// this uses the older MLLIB, with its RDD based interface
import org.apache.spark.mllib.linalg.Vectors
import org.apache.spark.mllib.regression.LabeledPoint

//class SurveySensorFull(user_id: Integer  ,timeslot: Long  ,survey_id: Integer  ,subsurvey_id: Integer  ,record_time: java.util.Date  ,device_id: String  ,timestamp: String  ,duration: Integer  ,resps: Map[Integer, String], activity: Double  ,x_axis_activity: Double,y_axis_activity: Double ,z_axis_activity: Double  ,orientationMag: Double  ,x_axis_orientation: Double  ,y_axis_orientation: Double  ,z_axis_orientation: Double  ,gyroscopeMag: Double  ,x_axis_gyroscope: Double  ,y_axis_gyroscope: Double ,z_axis_gyroscope: Double  ,features: org.apache.spark.ml.linalg.Vector,label: Double )


val dfJoinedSurveySensorSubset = dfJoinedSurveySensor.withColumn("wasAnsweredNumeric", answeredNotAnsweredNumericResponseCategoryUDF($"duration")).select("user_id", "timeslot", "survey_id", "subsurvey_id", "timestamp", "device_id","duration","activity","x_axis_activity","y_axis_activity","z_axis_activity","orientationMag","x_axis_orientation","y_axis_orientation","z_axis_orientation","gyroscopeMag","x_axis_gyroscope","y_axis_gyroscope","z_axis_gyroscope", "wasAnsweredNumeric")

val createLabeledPointUDF = udf((wasAnsweredNumeric:Double, activity:Double, orientationMag:Double, gyroscopeMag:Double, x_axis_orientation:Double, y_axis_orientation:Double, z_axis_orientation:Double) => LabeledPoint(wasAnsweredNumeric, Vectors.dense(activity, orientationMag, gyroscopeMag, x_axis_orientation, y_axis_orientation, z_axis_orientation)))

val dfJoinedSurveySensorSubsetWithLabel = dfJoinedSurveySensorSubset.withColumn("labeledpoint", createLabeledPointUDF($"wasAnsweredNumeric", $"activity", $"orientationMag", $"gyroscopeMag", $"x_axis_orientation", $"y_axis_orientation", $"z_axis_orientation"))

case class SurveySensorSelectedWithLabeled(user_id: Integer  ,timeslot: Long  ,survey_id: Integer  ,subsurvey_id: Integer   ,device_id: String  ,timestamp: String  ,duration: Integer,activity: Double  ,x_axis_activity: Double,y_axis_activity: Double ,z_axis_activity: Double  ,orientationMag: Double  ,x_axis_orientation: Double  ,y_axis_orientation: Double  ,z_axis_orientation: Double  ,gyroscopeMag: Double  ,x_axis_gyroscope: Double  ,y_axis_gyroscope: Double ,z_axis_gyroscope: Double, wasAnsweredNumeric: Double, labeledpoint:org.apache.spark.mllib.regression.LabeledPoint)

// convert to a typed dataset so that we can extract a typed RDD
val dsJoinedSurveySensorSubsetWithLabel = dfJoinedSurveySensorSubsetWithLabel.as[SurveySensorSelectedWithLabeled]

val rddLabeledPoints = dsJoinedSurveySensorSubsetWithLabel.map(_.labeledpoint).rdd
rddLabeledPoints.cache

// below is inspired by and adapted from code at https://spark.apache.org/docs/latest/mllib-linear-methods.html

import org.apache.spark.mllib.evaluation.BinaryClassificationMetrics
import org.apache.spark.mllib.util.MLUtils
import org.apache.spark.mllib.classification.SVMWithSGD
import org.apache.spark.mllib.classification.SVMModel
import org.apache.spark.rdd.RDD

val countIterations = 50

object SVMModelContainer  {
    def computeModel(rddLabeledPoints: RDD[LabeledPoint], countIterations:Int) =
        {
        val svmModel = SVMWithSGD.train(rddLabeledPoints, countIterations)
        svmModel.toString       // report on the model

        // ok, in order to evaluate the accuracy of prediction, we need to do a few things
        // firstly, alter it so that raw prediction scores are considered
        svmModel.clearThreshold
        // now, create a set of pairs -- in an RDD!
        val rddPredictionGroundTruthPairs = rddLabeledPoints.map(labeledPoint => (svmModel.predict(labeledPoint.features), labeledPoint.label))
        val metric = new BinaryClassificationMetrics(rddPredictionGroundTruthPairs);
        metric.areaUnderROC
        (svmModel, metric)
        }
     }

val (svmModel, classificationMetric) = SVMModelContainer.computeModel(rddLabeledPoints, countIterations)
classificationMetric.areaUnderROC
// 
