import twitter4j.TwitterFactory
import twitter4j.auth.AccessToken;
import twitter4j._
import twitter4j.Status;
import scala.collection.JavaConversions._
import scala.math._
import org.apache.spark.ml.PipelineModel;
import java.util.Date
import java.sql.Timestamp
import org.apache.spark.sql.functions._
import org.apache.spark.rdd._
import org.apache.spark.sql.SQLContext
import org.apache.spark.sql._


val sqlContext = new SQLContext(sc);
import sqlContext.implicits._;

val strHarvesterVersion = "14"


    // perform setting imperatively, as it appears to work better, for unknown reasons
    
def defineStandardGeoQuery(location:GeoLocation, radiusAround:Double, earliestIDProcessed:Long, maxAllowableTweetsInPage:Int) : Query = {
    val query= new Query()
    query.setGeoCode(location, radiusAround,Query.KILOMETERS)
    query.setSince("2016-12-09")
    query.setSinceId(earliestIDProcessed)
    query.setResultType(Query.RECENT)
    query.setCount(maxAllowableTweetsInPage)
    
    query
    }
    
// so that lines don't wrap, replace newlines by other whitespace (a space
// so that quotes within the tweet don't confuse parsing, replace double quotes by single quotes
//val encodeTweetStringForPersistenceUDF = udf((str:String) => "\"" + str.replaceAll("\r\n"," ").replaceAll("\n"," ").replaceAll("\"","'") + "\"")
val encodeTweetStringForPersistenceUDF = udf((str:String) => str.replaceAll("\r\n"," ").replaceAll("\n"," ").replaceAll("\"","'"))

case class EncodeableStatus(valueUnquoted:String, userId:Option[Long], user:Option[String], userLocation:Option[String], place: Option[String], latitude: Option[Double], longitude: Option[Double], timeCreated: java.sql.Timestamp, tweetId: Long, isRetweet:Boolean, approximateFavoriteCount:Int, tweetLanguage:Option[String], isUserGeoLocationEnabled:Option[Boolean])

def dataFrameForTweetStatuses(tweetsAsStatuses: Seq[Status]) : Dataset[EncodeableStatus] = 
{
    val encodeableTweetCollection = tweetsAsStatuses.map(status => EncodeableStatus(status.getText, 
                                                                                    Option(status.getUser).map(_.getId),
                                                                                    Option(status.getUser).map(_.getName),   // helps identify multiple tweets from same user, as opposed to an equal number of tweets from distinct users
                                                                                    Option(status.getUser).map(_.getLocation),   // helps identify place of tweet when not marked in tweet itself
                                                                                    
                                                                                    Option(status.getPlace).map(_.getFullName), 
                                                                                    Option(status.getGeoLocation).map(_.getLatitude), 
                                                                                    Option(status.getGeoLocation).map(_.getLongitude),  

                                                                                    new java.sql.Timestamp(status.getCreatedAt.getTime),
                                                                                    status.getId,
                                                                                    status.isRetweet,
                                                                                    status.getFavoriteCount,
                                                                                    Option(status.getLang),
                                                                                    Option(status.getUser).map(_.isGeoEnabled)
                                                                                    ))
    
    sc.parallelize(encodeableTweetCollection).toDS.withColumn("value", encodeTweetStringForPersistenceUDF($"valueUnquoted")).as[EncodeableStatus]
}

def saveDataframeToCSV(ds : Dataset[EncodeableStatus], filenameAndPath: String) 
    {     
    ds.select("user","userId", "userLocation", "place","latitude","longitude","timeCreated","tweetId", "isRetweet", "approximateFavoriteCount", "tweetLanguage", "isUserGeoLocationEnabled",  "value").withColumnRenamed("value", "text").repartition(1).write.mode(org.apache.spark.sql.SaveMode.Append).csv(filenameAndPath)
    } 

    
val consumerKey = "YOUR_INFO"
val consumerSecret = "YOUR_INFO"
val accessToken = "YOUR_INFO"
val accessTokenSecret = "YOUR_INFO"



val twitter = new TwitterFactory().getInstance()
  // Authorising with your Twitter Application credentials
twitter.setOAuthConsumer(consumerKey,consumerSecret)
twitter.setOAuthAccessToken(new AccessToken(accessToken, accessTokenSecret))


//def isPlausiblyILI(tweet: Status) : Boolean = { val strTweetText = tweet.getText;  strTweetText.contains("the flu") || strTweetText.contains("influenza") || strTweetText.contains("a cold") || strTweetText.contains("bad cold") || strTweetText.contains("a flu") || strTweetText.contains("sneez") || strTweetText.contains("cough") || strTweetText.contains("body ache") || strTweetText.contains("fever") }


val logisticRegressionModelFromCrossValidation = PipelineModel.load("output/HealthRecognitionModel_Dec-10-2016")
    



// the query is updated using the paging associated with the cursor mechanism
//var query = new Query().query("sick OR ill").geoCode(new GeoLocation(42.3601, -71.0), 100, "mi").since("2016-11-10").count(maxAllowableTweetsInPageForAPI)

// Use: val (countTweetsProcessed, countInterestingTweetsProcessed, interestingTweets) = identifyInterestingTweetsRecursive(10000, isPlausiblyILI _ ,new Query().geoCode(new GeoLocation(52.1332, -106.67), 1000,"km").count(100))

// val (countTweetsProcessedSince1120, countInterestingTweetsProcessedSince1120, interestingTweetsSince1120) = identifyInterestingTweetsRecursive(10000, isPlausiblyILI _ ,new Query().geoCode(new GeoLocation(52.1332, -106.67), 1000,"km").count(100).since("2016-11-20"))


//val (countTweetsProcessed2, countInterestingTweetsProcessed2, interestingTweets2, earliestDateProcessed) = identifyInterestingTweetsRecursive(10000, isPlausiblyILI _ ,new Query().geoCode(new GeoLocation(52.1332, -106.67), 1000,"km").resultType(Query.RECENT).count(maxAllowableTweetsInPageForAPI))
// the results go back much further if use a search query (rather than just no search query)

//val (countTweetsProcessed7, countInterestingTweetsProcessed7, interestingTweets7, earliestDateProcessed7) = identifyInterestingTweetsRecursive(10000, isPlausiblyILI _ ,new Query().geoCode(new GeoLocation(52.1332, -106.67), 100,"km").query(""""flu" OR "sick" OR "ill"""").since("2016-11-20").resultType(Query.RECENT).count(maxAllowableTweetsInPageForAPI))
//val (countTweetsProcessed5, countInterestingTweetsProcessed5, interestingTweets5, earliestDateProcessed5) = identifyInterestingTweetsRecursive(10000, isPlausiblyILI _ ,new Query().geoCode(new GeoLocation(52.1332, -106.67), 100,"km").query("sick").since("2016-11-20").resultType(Query.RECENT).count(maxAllowableTweetsInPageForAPI))

// The following method takes
//      a maximum count of tweets to examine (process)
//      a minimum id to harvest (important because it does not appear that the Twitter API reliably screens out those beyond a certain specified value, despite the querySinceId option)
//      a predicate taking a Status associated with a tweet, and returning a boolean indicating if that tweet is of interest to return
//      a query to execute
// returns a tuple of
//      the total number of tweets processed (whether interesting or not, but ignoring ones encountered with ids < minIdToHarvest)
//      the total number of interesting tweets found
//      a  Sequence of those that are interesting 
//      the earliest Date actually processed within the above
//      the earliest ID actually processed within the above
//      the LATEST ID encountered and processed within the above
//      the LATEST date encountered and processed within the above
def identifyInterestingTweetsRecursive(maxTweetsToProcess: Int, minIdToHarvest:Long, fnIsInteresting: (Status => Boolean), query: Query) : (Int, Int, Seq[Status], Option[java.util.Date], Option[Long], Option[Long], Option[java.util.Date]) = 
    {
    val results = twitter.search(query)
    // get the total set of tweets in this page -- putting aside the issue as to whether we will process them (or not because we have already seen them)
    // note that the semantics of take are such that it returns a collection of tweets whose cardinality is the minimum of 
    // countTweetsRemainingUpperBound and the number of tweets in this Page
    val tweetsInPage = results.getTweets.take(maxTweetsToProcess)

    // get the earliest id in this page -- regardless as to whether we will go back that far (i.e., whether or not it has previously been seen)
    val earliestIdThisPage = tweetsInPage.last.getId // this is important to figure out if we have to go back any further
   
   // now compute the subset of tweets in the page that we will seek to process
   // note that tweetsInPageToProcess will include the same tweets as tweetsInPage 
   //   if we haven't seen any of these before (i.e., if we haven't yet reached our minIdToHarvest)
    val tweetsInPageToProcess = tweetsInPage.filter((status:Status) => status.getId >= minIdToHarvest)      
    
    // get the number of tweets to process in this Page, considering 
    //      the number available
    //      whether or not we will in fact process them (taking into account whether we have previously seen them) 
    //      the number we still need to process
    val countTweetsProcessedThisPage = tweetsInPageToProcess.size

    // ok, we have reached a page consisting entirely of Ids < minIdToHarvest.  
    if (countTweetsProcessedThisPage == 0)
        {
        // In this case, we have 0 processed values and don't have any earliest date encountered, earliest id encountered, or latest id encountered
        ( 0, 0, Seq(), None, None, None, None)
        }
    else
        {
        val earliestIdToProcessThisPage = tweetsInPageToProcess.last.getId

        val latestIdThisPage = tweetsInPageToProcess.head.getId 
        
        val latestDateThisPage = tweetsInPageToProcess.head.getCreatedAt

        val earliestDateThisPage = tweetsInPageToProcess.last.getCreatedAt
        
        //    ok, now filter the tweets to be specifically the interesting one
        val interestingTweetsThisPage = tweetsInPageToProcess.filter((status:Status) => fnIsInteresting(status)).toVector
        val countInterestingTweetsThisPage = interestingTweetsThisPage.size

        println(s"$countTweetsProcessedThisPage tweets received on page ($countInterestingTweetsThisPage of possible interest)");
        // we need to continue only if we have BOTH additional tweets remaining and tweets that we still need to process
        // Please note that countTweetsProcessedThisPage will only be different from tweetsInPage.size if earliestIdThisPage <= minIdToHarvest
        if ((countTweetsProcessedThisPage < maxTweetsToProcess) && earliestIdThisPage > minIdToHarvest && results.hasNext)
            {
            val (countTweetsProcessedInLaterPages, countInterestingTweetsProcessedInLaterPages, interestingTweetsInLaterPages, earliestDateProcessed, earliestIDProcessed, latestIDProcessedEarlierPages, latestDateProcessedEarlierPages) = identifyInterestingTweetsRecursive(maxTweetsToProcess - countTweetsProcessedThisPage, minIdToHarvest, fnIsInteresting, results.nextQuery)
            
            // return a 3-tuple combining the results in this page with those from later pages
            ( countTweetsProcessedThisPage + countTweetsProcessedInLaterPages, 
            countInterestingTweetsThisPage + countInterestingTweetsProcessedInLaterPages,
            interestingTweetsThisPage ++ interestingTweetsInLaterPages,     // if there was nothing to process on the later pages (because we already had seen them all), interestingTweetsInLaterPages will be an empty sequence
            Some(earliestDateProcessed.getOrElse(earliestDateThisPage)),          // if there was nothing to process on the later pages (because we already had seen them all), just use the earliest date processed from this page
            Some(earliestIDProcessed.getOrElse(earliestIdToProcessThisPage)),     // if there was nothing to process on the later pages (because we already had seen them all), just use the earliest date processed from this page
            Some(latestIdThisPage),                                               // certainly exists -- we would have left earlier if not
            Some(latestDateThisPage)                                              // certainly exists -- we would have left earlier if not
            )
            }
        else
            // otherwise, if this is the final page, just return the number from this page
            ( countTweetsProcessedThisPage, countInterestingTweetsThisPage, interestingTweetsThisPage, Some(earliestDateThisPage), Some(earliestIdToProcessThisPage), Some(latestIdThisPage), Some(latestDateThisPage) )
        }
    }

// search the area around Boston
val maxAllowableTweetsInPageForAPI = 100

// The below is accepted -- but appears to bring in locations from outside of the area (?)
//val (countTweetsProcessed7, countInterestingTweetsProcessed7, interestingTweets7, earliestDateProcessed7) = identifyInterestingTweetsRecursive(10000, (status:Status) => true ,new Query().geoCode(new GeoLocation(52.1332, -106.67), 350,"km").since("2016-12-09").resultType(Query.RECENT).count(maxAllowableTweetsInPageForAPI))

//val (countTweetsProcessed9, countInterestingTweetsProcessed9, interestingTweets9, earliestDateProcessed9) = identifyInterestingTweetsRecursive(10000, (status:Status) => true ,new Query().geoCode(new GeoLocation(52.1332, -106.67), 350,Query.Unit.valueOf("km")).since("2016-12-09").resultType(Query.RECENT).count(maxAllowableTweetsInPageForAPI))

var latestIDProcessedSouthern:Long = -1
var latestIDProcessedNorthern:Long = -1




val countMaxTweetsToHarvestPerBatch =  7000

val countMinutesBetweenHarvests = 20    // to avoid running into race conditions involving twitter's 15 minute limit
val countSecondsBetweenHarvests = 60 * countMinutesBetweenHarvests

// ok, now loop every countMinutesBetweenHarvests and harvest from twitter.  
// NB: it is advised to make countMinutesBetweenHarvests > 15 to avoid exceeding twitter 
// limits
    

while (true)
    {
    val startingIdForNextBatchSouthern = latestIDProcessedSouthern + 1;
    val startingIdForNextBatchNorthern = latestIDProcessedNorthern + 1;
    
    val querySouthernSK = defineStandardGeoQuery(new GeoLocation(51.591803, -105.688616), 425, startingIdForNextBatchSouthern, maxAllowableTweetsInPageForAPI)
    val queryNorthernSK = defineStandardGeoQuery(new GeoLocation(57.144607, -105.871954), 400, startingIdForNextBatchNorthern, maxAllowableTweetsInPageForAPI)

    val (countTweetsProcessedSouthern, countInterestingTweetsProcessedSouthern, allTweetsSouthern, earliestDateProcessedThisBatchSouthern, earliestIDProcessedThisBatchSouthern, latestIDProcessedThisBatchSouthern, latestDateProcessedThisBatchSouthern) = identifyInterestingTweetsRecursive(countMaxTweetsToHarvestPerBatch, startingIdForNextBatchSouthern, (status:Status) => true, querySouthernSK)
    latestIDProcessedSouthern = latestIDProcessedThisBatchSouthern.getOrElse(latestIDProcessedSouthern)         // there is a possibility that there are no new tweets
    if (countTweetsProcessedSouthern > 0)
        println(s"Processed $countTweetsProcessedSouthern tweets from southern SK; earliest date, id processed was ($earliestDateProcessedThisBatchSouthern, $earliestIDProcessedThisBatchSouthern), final (date, id) processed is ($latestDateProcessedThisBatchSouthern, $latestIDProcessedSouthern)")
    else
        println("No tweets available from southern SK");
        
    val (countTweetsProcessedNorthern, countInterestingTweetsProcessedNorthern, allTweetsNorthern, earliestDateProcessedThisBatchNorthern, earliestIDProcessedThisBatchNorthern, latestIDProcessedThisBatchNorthern, latestDateProcessedThisBatchNorthern) = identifyInterestingTweetsRecursive(countMaxTweetsToHarvestPerBatch, startingIdForNextBatchNorthern, (status:Status) => true, queryNorthernSK)
    latestIDProcessedNorthern = latestIDProcessedThisBatchNorthern.getOrElse(latestIDProcessedNorthern)
    if (countTweetsProcessedNorthern > 0)
        println(s"Processed $countTweetsProcessedNorthern tweets from northern SK; earliest date, id encountered was ($earliestDateProcessedThisBatchNorthern, $earliestIDProcessedThisBatchNorthern), final (date, id) processed is ($latestDateProcessedThisBatchNorthern, $latestIDProcessedNorthern)")
    else
        println("No tweets available from northern SK");
        
    // these appear to be better grounded tweets
    // we are operating here off of collections of status reports -- not yet dataframes
    
    val dsTweetsSouthern = dataFrameForTweetStatuses(allTweetsSouthern)
    val dsTweetsNorthern = dataFrameForTweetStatuses(allTweetsNorthern)

    // because the tweets may have some overlap, make them distinct
    val dsTweets = dsTweetsSouthern.union(dsTweetsNorthern).distinct
    saveDataframeToCSV(dsTweets, s"AllSKTweets/$strHarvesterVersion")

    Thread.sleep(countSecondsBetweenHarvests * 1000)
    }
    
//val (countTweetsProcessed8, countInterestingTweetsProcessed8, interestingTweets8, earliestDateProcessed8) = identifyInterestingTweetsRecursive(200, (status:Status) => true ,new GeoQuery(new GeoLocation(52.1332, -106.67)).accuracy("350000").since("2016-12-09").resultType(Query.RECENT).count(maxAllowableTweetsInPageForAPI))

// Because the twitter search API doesn't include support for the "near" operator, the below fails
//val (countTweetsProcessed7, countInterestingTweetsProcessed7, interestingTweets7, earliestDateProcessed7) = identifyInterestingTweetsRecursive(10000, (status:Status) => true ,new Query().query("""near:"Saskatchewan"""").since("2016-11-20").resultType(Query.RECENT).count(maxAllowableTweetsInPageForAPI))

