// ok, go through the standard setup 
sc.stop()
import com.datastax.spark.connector._
import org.apache.spark.SparkContext
import org.apache.spark.SparkContext._
import org.apache.spark.SparkConf
val conf = new SparkConf(true).set("spark.cassandra.connection.host", "blinchiki.usask.ca").set("spark.cassandra.auth.username", "YOUR_USERNAME").set("spark.cassandra.auth.password", "YOUR_PASSWORD")
val sc = new SparkContext("local", "CassandraSurveyAccess", conf)
import org.apache.spark.sql.SQLContext
val sqlContext = new SQLContext(sc)
import sqlContext.implicits._

// a reminder of the steps for accessing sensor data
//  create a dataframe for the whole of the GPS table
val dfGPS = sqlContext.read.format("org.apache.spark.sql.cassandra").options(Map("keyspace" -> "ethica_study_73", "table" -> "gps")).load() 

// examine the schema
dfGPS.printSchema

// recall that can filter the data (e.g., based on accuracy) and aggregate by count
dfGPS.filter($"accu" < 20.0).groupBy("provider").count.show
// a faster way of performing the above
dfGPS.filter($"accu" < 20.0).groupBy("provider").count.show

// ok, we now get the survey responses
val dfSurveyResponses = sqlContext.read.format("org.apache.spark.sql.cassandra").options(Map("keyspace" -> "ethica_study_73", "table" -> "survey_responses")).load() 
dfSurveyResponses.printSchema
dfSurveyResponses.cache
dfSurveyResponses.groupBy($"survey_id", $"subsurvey_id").count.show
// show the count responses to various surveys_id
dfSurveyResponses.groupBy($"survey_id").count.show(100)
// ok, order by the surveys with the highest number of survey responses
dfSurveyResponses.groupBy($"survey_id").count.orderBy(-$"count(1)").show(100)
// a more transparent way to do this
dfSurveyResponses.groupBy($"survey_id").count.orderBy($"count(1)".desc).show(100)
// count the number of responses, expirations and cancellations to survey 242 associated with different users
dfSurveyResponses.filter($"survey_id" === 242).groupBy($"user_id").count.show
// ok now examine the different durations
dfSurveyResponses.filter($"survey_id" === 242).groupBy($"duration").count.show
// ok, show ascending counts
dfSurveyResponses.filter($"survey_id" === 242).groupBy($"duration").count.orderBy($"duration").show(200)
// ok, break the survey actually responsed into 5 minute duration bins (omitting cancellations, expirations)
dfSurveyResponses.filter($"survey_id" === 242).filter($"duration" >= 0).groupBy(round($"duration"/5)*5).count.orderBy("(round((duration / 5), 0) * 5)").show(200)
// count the number of distinct subsurveys
dfSurveyResponses.filter($"survey_id" === 242).select($"subsurvey_id").distinct.count

// looking at the responses.  these are JSON data
dfSurveyResponses.filter($"survey_id" === 242).select("resps").show
// render the responses into arrays
dfSurveyResponses.filter($"survey_id" === 242).select("resps").collect
// looking at the first in greater detail -- note the Map, each of whose entries has an integer key (staring at 1) and a value that represents one or more responses to each question
// (there can be more than one such response to a question if, for example, the question allows for multiple selections)
dfSurveyResponses.filter($"survey_id" === 242).select("resps").collect.head
// ok demonstration of how to obtain the responses to the first question
// Please note that we are here getting the map associated with the only column (named "resps")
dfSurveyResponses.filter($"survey_id" === 242).select("resps").map(row => row.getMap(row.fieldIndex("resps")).getOrElse(1, "NA")).show
// ok, the below retrieves the the responses to the first question (question 1) as a scala array
dfSurveyResponses.filter($"survey_id" === 242).select("resps").map(row => row.getMap(row.fieldIndex("resps")).getOrElse(1, "NA")).collect
// ok, the below retrieves the the responses to the third question (question 3) as a scala array
dfSurveyResponses.filter($"survey_id" === 242).select("resps").map(row => row.getMap(row.fieldIndex("resps")).getOrElse(3, "NA")).collect
// the below provides a way to print this response
dfSurveyResponses.filter($"survey_id" === 242).select("resps").map(row => row.getMap(row.fieldIndex("resps")).getOrElse(3, "NA")).foreach(println(_))
// ok, now only consider actual responses to the survey
dfSurveyResponses.filter($"survey_id" === 242).filter($"user_id" === 231).filter($"duration" >= 0).show
// count the number
dfSurveyResponses.filter($"survey_id" === 242).filter($"user_id" === 231).filter($"duration" >= 0).count
// create a user defined function that both
//  a) Abstracts this notion of "response" to make it more modular and transparent
//  b) Lifts the function from operating on values to operate on columns
val isRespondedToUDF = udf((duration:Integer) => (duration >= 0))
dfSurveyResponses.filter($"survey_id" === 242).filter($"user_id" === 231).filter($"duration" >= 0).count
// count the number of surveys having a response duration (or cancellation/expiration) falling in certain values
dfSurveyResponses.filter($"survey_id" === 242).filter($"user_id" === 231).groupBy($"duration").count.show
// ok, now do grouping as to whether the survey was reponded to
dfSurveyResponses.filter($"survey_id" === 242).filter($"user_id" === 231).groupBy(isRespondedToUDF($"duration")).count.withColumnRenamed("UDF(duration)", "isAnswered").show

// a preview of the next class: getting particular values
dfSurveyResponses.filter($"survey_id" === 242).select("resps").map(row => row.getMap(row.fieldIndex("resps")).getOrElse(1, "NA")).collect


