// ok, go through the standard setup 
sc.stop()
import com.datastax.spark.connector._
import org.apache.spark.SparkContext
import org.apache.spark.SparkContext._
import org.apache.spark.SparkConf
val conf = new SparkConf(true).set("spark.cassandra.connection.host", "blinchiki.usask.ca").set("spark.cassandra.auth.username", "YOUR_USERNAME").set("spark.cassandra.auth.password", "YOUR_PASSWORD")
val sc = new SparkContext("local", "test", conf)
import org.apache.spark.sql.SQLContext
val sqlContext = new SQLContext(sc);
import sqlContext.implicits._;

// first we work to divide the survey data into timeslots (time bins)
val dfSurveyResponses = sqlContext.read.format("org.apache.spark.sql.cassandra").options(Map("keyspace" -> "ethica_study_73", "table" -> "survey_responses")).load()
val dfPASamplingSurveyResponses = dfSurveyResponses.filter($"user_id" === 231).filter($"survey_id" === 242)
dfPASamplingSurveyResponses.printSchema
// ok, now we need to group the time into time bins, so we can use that as a key
// NB: here we choose a 10 minute timeslot (the "10" in the formula below), but you may wish to impose timeslots of different length
val timeslotFromTimestampUDF = udf((ts: java.sql.Timestamp) => ts.getTime() / (10*60*1000))
val dfTimeslottedSurveyResponses = dfPASamplingSurveyResponses.withColumn("timeslot", timeslotFromTimestampUDF($"record_time"))
// ok, for convenience, we narrow this to fewer columns, since we don't need all columns
// note that for other uses, may we wish to also include the "resps" column
val dfTimeslottedSurveyResponsesNarrowed = dfTimeslottedSurveyResponses.select("timeslot", "user_id", "duration")
dfTimeslottedSurveyResponsesNarrowed.printSchema
dfTimeslottedSurveyResponsesNarrowed.cache
dfTimeslottedSurveyResponsesNarrowed.show(20)
// explain the origin of this dataframe
dfTimeslottedSurveyResponsesNarrowed.explain()

// ok, having situated the survey data in clear timeslots above, we go on to aggregate the accelerometer data into timeslots
val dfAccelVar = sqlContext.read.format("org.apache.spark.sql.cassandra").options(Map("keyspace" -> "ethica_study_73", "table" -> "accel_variability")).load()
dfAccelVar.printSchema
// the below suggests that this is not loaded into memory
dfAccelVar.explain()
// use the udf defined above to find the timeslot for the responses
// and create a column with the time of the accelerometer data binned into timeslot
val dfTimeslottedAccelVar = dfAccelVar.withColumn("timeslot", timeslotFromTimestampUDF($"datetime_bin"))
// ok, now group the accelerometer by TIMESLOT and USER_ID
val dfTimeslotAccelVarKV = dfTimeslottedAccelVar.groupBy("timeslot", "user_id").agg(avg(sqrt($"avg_x"*$"avg_x" + $"avg_y"*$"avg_y" + $"avg_z"*$"avg_z")))
// a better way of doing this (more transparent and more reuseable)
import scala.math.sqrt
val normUDF = udf((x:Float, y:Float, z:Float) => sqrt((x*x + y*y + z*z).toDouble))
// ok, now aggregate the accel data in the same way (into timeslot)
val dfTimeslotAccelVarKV = dfTimeslottedAccelVar.groupBy("timeslot", "user_id").agg(avg(normUDF($"avg_x", $"avg_y", $"avg_z"))).withColumnRenamed("avg(UDF(avg_x, avg_y, avg_z))","activity")
dfTimeslotAccelVarKV.printSchema
dfTimeslotAccelVarKV.cache()
dfTimeslotAccelVarKV.explain()
dfTimeslotAccelVarKV.printSchema
// a function to bin the activity into categories (each of size 1)
// NB: for other uses, one may wish to use finer-grained bins (e.g., of size 0.25)
val activityBinsUDF = udf((activity:Double) => math.round(activity))
// try looking at the count of timeslot values that fall into each of these bins
dfTimeslotAccelVarKV.groupBy(activityBinsUDF($"activity")).count.orderBy("UDF(activity)").show

// ok, now the below joins together the survey and accelerometer data
// ok, now join the accelerometer data with the survey response data
val joinedSurveyAccelVar = dfTimeslottedSurveyResponsesNarrowed.join(dfTimeslotAccelVarKV, usingColumns = List("user_id", "timeslot"))
joinedSurveyAccelVar.cache
// the below gives a breakdown, but has too fine-grained a set of durations
joinedSurveyAccelVar.groupBy(activityBinsUDF($"activity"), $"duration").count.show
val responseCategoryUDF = udf((duration:Int) => (duration match { case -2 => "Cancelled"; case -1 => "Expired"; case n if (n >= 0) => "Answered"; case m => "Unknown: " + m; }))
// the below groups into duration categories
joinedSurveyAccelVar.groupBy(activityBinsUDF($"activity"), responseCategoryUDF($"duration")).count.show
// now show the new columns
joinedSurveyAccelVar.withColumn("responseCategory", responseCategoryUDF($"duration")).withColumn("activityBin", activityBinsUDF($"activity")).printSchema

// ok, now create a crosstab of activity by bins
joinedSurveyAccelVar.withColumn("responseCategory", responseCategoryUDF($"duration")).withColumn("activityBin", activityBinsUDF($"activity")).stat.crosstab("activityBin", "responseCategory").printSchema
// Please note that the ordering of the above is not very effective
// we now perform an ordering
// unfortunately, that ordering is LEXICOGRAPHIC, rather than numeric!
joinedSurveyAccelVar.withColumn("responseCategory", responseCategoryUDF($"duration")).withColumn("activityBin", activityBinsUDF($"activity")).stat.crosstab("activityBin", "responseCategory").orderBy("activityBin_responseCategory").show
// now see how the activity levels vary by surveys the have been answered vs. those which have not been answered
joinedSurveyAccelVar.groupBy(responseCategoryUDF($"duration")).agg(avg($"activity")).show
