// perform the standard loading of the requisite libraries
sc.stop()
import com.datastax.spark.connector._
import org.apache.spark.SparkContext
import org.apache.spark.SparkContext._
import org.apache.spark.SparkConf
val conf = new SparkConf(true).set("spark.cassandra.connection.host", "blinchiki.usask.ca").set("spark.cassandra.auth.username", "YOUR_USERNAME").set("spark.cassandra.auth.password", "YOUR_PASSWORD")
val sc = new SparkContext("local", "test", conf)
import org.apache.spark.sql.SQLContext
val sqlContext = new SQLContext(sc);
import sqlContext.implicits._;

//************Begin standard cross-linking of sensor and survey data************
// now also load the math libraries so that we can use sqrt when computing the norm
import scala.math.sqrt

// ok, the below just goes through the standard procedures for 
//  a) loading a subset of survey responses
//  b) loading the accelerometer data
//  c) joining the survey responses with the the accelerometer data
val dfSurveyResponses = sqlContext.read.format("org.apache.spark.sql.cassandra").options(Map("keyspace" -> "ethica_study_73", "table" -> "survey_responses")).load()
val dfPASamplingSurveyResponses = dfSurveyResponses.filter($"user_id" === 231).filter($"survey_id" === 242)
// ok, now we need to group the time into time bins, so we can use that as a key
// NB: here we choose a 10 minute timeslot (the "10" in the formula below), but you may wish to impose timeslots of different length
val timeslotFromTimestampUDF = udf((ts: java.sql.Timestamp) => ts.getTime() / (10*60*1000))
val dfTimeslottedSurveyResponses = dfPASamplingSurveyResponses.withColumn("timeslot", timeslotFromTimestampUDF($"record_time"))
val dfTimeslottedSurveyResponsesNarrowed = dfTimeslottedSurveyResponses.select("timeslot", "user_id", "duration", "resps")

// ok, with the survey data in hand, now process the accelerometer data
val dfAccelVar = sqlContext.read.format("org.apache.spark.sql.cassandra").options(Map("keyspace" -> "ethica_study_73", "table" -> "accel_variability")).load()
// use the udf defined above to find the timeslot for the responses
// and create a column with the time of the accelerometer data binned into timeslot
val dfTimeslottedAccelVar = dfAccelVar.withColumn("timeslot", timeslotFromTimestampUDF($"datetime_bin"))
// ok, now group the accelerometer by TIMESLOT and USER_ID
// a better way of doing this (more transparent and more reuseable)

val normUDF = udf((x:Float, y:Float, z:Float) => sqrt((x*x + y*y + z*z).toDouble))
// ok, now aggregate the accel data in the same way (into timeslot)
val dfTimeslottedAccelVarKV = dfTimeslottedAccelVar.groupBy("timeslot", "user_id").agg(avg(normUDF($"avg_x", $"avg_y", $"avg_z"))as("activity"))
dfTimeslottedAccelVarKV.cache()

// ok, now the below joins together the survey and accelerometer data
// ok, now join the accelerometer data with the survey response data
val joinedSurveyAccelVar = dfTimeslottedSurveyResponsesNarrowed.join(dfTimeslottedAccelVarKV, usingColumns = List("user_id", "timeslot"))

//************End standard cross-linking of sensor and survey data************
//************Below are the components associated with questions responses************


// Ok, the above was just the standard routine to cross-link the survey and accelerometer data
// We transition below to depending on the question responses
val extractQuestionNumberJSONUDFGeneratorForQuestionNumber = ((iDesiredQuestion:Int) => udf((m:Map[Int,String]) => m.getOrElse(iDesiredQuestion,"")))
// we'd like to do something like the next line -- assigning unique row ids to each row, so that we can later join the extracted question answer data for each row (extracted from json) with the
// dataframe for each row, but because we can't join in such row information with the extracted information, it does not currently seem like a viable option
//val joinedSurveyAccelVarActuallyRespondedToWithQuestionInfo = joinedSurveyAccelVar.filter($"duration" >= 0).withColumn("rowId", monotonically_increasing_id).withColumn("q1Info", 
val joinedSurveyAccelVarActuallyRespondedToWithQuestionInfo = joinedSurveyAccelVar.filter($"duration" >= 0).withColumn("q1Info", extractQuestionNumberJSONUDFGeneratorForQuestionNumber(1)($"resps")).withColumn("q3Info", extractQuestionNumberJSONUDFGeneratorForQuestionNumber(3)($"resps"))

// A cleaner way for handling questions content and responses is shown below -- we parse the json directly using a call to the lift json library 
import net.liftweb.json._
implicit val formats = DefaultFormats

// create case classes to support the data structures
// Please note that while resp_time is a structured value, it doesn't appear to be parsable by java.time or the older standard library for java time.  We thus leave it as a string
case class EthicaSurveyQuestionSingleResponse(resp_time: String, answer_content:String, answer_id:Long)
case class EthicaSurveyResponse(q_content: String, resp:Array[EthicaSurveyQuestionSingleResponse], q_type:String, version:String)

// create a user defined function to parse a json field for a response to extract the response TEXT
val parseSingleAnswerJSONResponseContentUDF = udf((strEthicaQuestionResponse:String) => { implicit val formats = DefaultFormats; parse(strEthicaQuestionResponse).extract[EthicaSurveyResponse].resp(0).answer_content }) 
// demonstration of how things works
joinedSurveyAccelVarActuallyRespondedToWithQuestionInfo.withColumn("response", parseSingleAnswerJSONResponseContentUDF($"q1Info")).show

// create a user defined function to parse a json field for a response to extract the response ID
val parseSingleAnswerJSONResponseIdUDF = udf((strEthicaQuestionResponse:String) => { implicit val formats = DefaultFormats; parse(strEthicaQuestionResponse).extract[EthicaSurveyResponse].resp(0).answer_id }) 

// create a user defined function to return all of the items listed as a commma separated list
val extractAnswerTextsUDF = udf((strEthicaQuestionResponse:String) => { implicit val formats = DefaultFormats; if (strEthicaQuestionResponse.length == 0) "" else parse(strEthicaQuestionResponse).extract[EthicaSurveyResponse].resp.map(specificResponse => specificResponse.answer_content).mkString(",") }) 

// include both answer ids and answer text
val joinedSurveyAccelVarActuallyRespondedToWithQ1Answers = joinedSurveyAccelVarActuallyRespondedToWithQuestionInfo.withColumn("q1_answer_text", parseSingleAnswerJSONResponseContentUDF($"q1Info")).withColumn("q1_answer_id", parseSingleAnswerJSONResponseIdUDF($"q1Info"))
joinedSurveyAccelVarActuallyRespondedToWithQ1Answers.show

// create a user defined function to return a boolean indicating if a given answerId is included among the answers given
val parseMultiAnswerJSONResponseIncludesIdTestUDFGeneratorForAnswerId = ((answerId:Long) => udf((strEthicaQuestionResponse:String) => { implicit val formats = DefaultFormats; if (strEthicaQuestionResponse.length == 0) false else parse(strEthicaQuestionResponse).extract[EthicaSurveyResponse].resp.exists(specificResponse => specificResponse.answer_id == answerId) })) 

// create a user defined function to return a boolean indicating if a given answer content contains the text given
val parseMultiAnswerJSONResponseContainsTextTestUDFGeneratorForAnswerText = ((strWhosePresenceToSearch:String) => udf((strEthicaQuestionResponse:String) => { implicit val formats = DefaultFormats; if (strEthicaQuestionResponse.length == 0) false else parse(strEthicaQuestionResponse).extract[EthicaSurveyResponse].resp.exists(specificResponse => specificResponse.answer_content.contains(strWhosePresenceToSearch) ) })) 

val joinedSurveyAccelVarActuallyRespondedToWithQ3Answers = joinedSurveyAccelVarActuallyRespondedToWithQ1Answers.withColumn("q3_IsAnswer1Present", parseMultiAnswerJSONResponseIncludesIdTestUDFGeneratorForAnswerId(10)($"q3Info")).withColumn("q3_IsRunningPresent", parseMultiAnswerJSONResponseContainsTextTestUDFGeneratorForAnswerText("Running")($"q3Info")).withColumn("q3_answers", extractAnswerTextsUDF($"q3Info"))

joinedSurveyAccelVarActuallyRespondedToWithQ3Answers.show

// we now show some calculations involving the results here

joinedSurveyAccelVarActuallyRespondedToWithQ3Answers.groupBy($"q1_answer_id").count.show

joinedSurveyAccelVarActuallyRespondedToWithQ3Answers.groupBy($"q1_answer_id").agg(avg("activity")).show

joinedSurveyAccelVarActuallyRespondedToWithQ3Answers.groupBy($"q1_answer_text").agg(avg("activity")).show
